/************************************************************************/
/*
  dynamics.c: This is where the numerical integration is done.
*/
/************************************************************************/
/*
In this version we are going to use the following conventions
x forward
y left
z up
arm straight down is zero.
angles are positive rotating around the -y axis.
*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "main.h"
#include "main2.h"

/************************************************************************/
/* GLOBAL VARIABLES */

/* Simulation structure */
SIM sim;

/************************************************************************/
/************************************************************************/

#ifdef WIN32
int isnan( double x )
{ return 0; }

int isinf( double x )
{ return 0; }
#endif

/************************************************************************/

static void forward_kinematics( SIM *s )
{
  int i;

  // joint locations
  for( i = 0; i < 3; i++ )
    s->shoulder[i] = 0;
    
  s->elbow[XX] = s->upper_arm_length*sin( s->angle[SHOULDER] )
    + s->shoulder[XX];
  s->elbow[YY] = s->shoulder[YY];
  s->elbow[ZZ] = -s->upper_arm_length*cos( s->angle[SHOULDER] )
    + s->shoulder[ZZ];

  s->wrist[XX] = s->forearm_length*sin( s->angle[SHOULDER] + s->angle[ELBOW] )
    + s->elbow[XX];
  s->wrist[YY] = s->elbow[YY];
  s->wrist[ZZ] = -s->forearm_length*cos( s->angle[SHOULDER] + s->angle[ELBOW] )
    + s->elbow[ZZ];
}

/**********************************************************************/
/*
This needs to be called whenever the dynamics inertial parameters are changed.
*/

void set_dynamics_constants( SIM *s )
{
  double m0 = s->upper_arm_mass;
  double m1 = s->forearm_mass;
  double l0cm = s->upper_arm_cm;
  double l1cm = s->forearm_cm;
  double l0 = s->upper_arm_length;
  double I0yy = s->upper_arm_I; // moment of inertia about COM
  double I1yy = s->forearm_I; // moment of inertia about COM

  s->total_mass = m0 + m1;
  s->l0cm_m0 = l0cm*m0;
  s->l1cm_m1 = l1cm*m1;
  s->g_l1cm_m1 = s->gravity*l1cm*m1;
  s->l1cm_l1cm_m1 = l1cm*s->l1cm_m1;
  s->l0_l1cm_m1 = l0*s->l1cm_m1;
  s->l0cm_m0_p_l0_m1 = s->l0cm_m0 + l0*m1;
  s->g_l0cm_m0_p_l0_m1 = s->gravity*s->l0cm_m0_p_l0_m1;
  s->expr2 = I1yy + s->l1cm_l1cm_m1;
  s->expr3 = I0yy + l0cm*s->l0cm_m0 + l0*l0*m1;
}

/**********************************************************************/

int set_state( SIM *s, double time, double *state )
{
  int i; 

  s->time = time;
  for ( i = 0; i < N_X; i++ )
    s->state[i] = state[i];

  forward_kinematics( s );

  // just for good luck
  set_dynamics_constants( s );

  return 1;
}

/************************************************************************/

void init_dynamics( SIM *s )
{

  s->status = OK;
  forward_kinematics( s );
}

/**********************************************************************/

void forward_dynamics( SIM *s, double *action, double *accelerations )
{
  double a, b, c, d, e, f;
  double determinant;
  double sa0, ca0, sa1, ca1;
  double a0d;
  double a1d;
  double a0dd;
  double a1dd;
  double a0d_2, a0d_a1d, a1d_2;
  double sa0a1, ca0a1;
  double a0d_p_a1d_2;
  double l1cm_m1_sa0a1, l1cm_m1_ca0a1;
  double l0_l1cm_m1_sa1, l0_l1cm_m1_ca1;
  double l0cm_m0_p_l0_m1_sa0, l0cm_m0_p_l0_m1_ca0;
  double expr1;

  /* These dynamics were derived for straight up to be the zero position,
     so I am modifying them here so straight down is the zero position
     by subptracting PI from the shoulder angle.
  */
  sa0 = sin( s->angle[SHOULDER] - M_PI );
  ca0 = cos( s->angle[SHOULDER] - M_PI );
  sa1 = sin( s->angle[ELBOW] );
  ca1 = cos( s->angle[ELBOW] );
  sa0a1 = sa0*ca1 + ca0*sa1;
  ca0a1 = ca0*ca1 - sa0*sa1;

  a0d = s->angled[SHOULDER];
  a1d = s->angled[ELBOW];
  a0d_2 = a0d*a0d;
  a0d_a1d = a0d*a1d;
  a1d_2 = a1d*a1d;
  a0d_p_a1d_2 = (a0d + a1d)*(a0d + a1d);

  l1cm_m1_sa0a1 = s->l1cm_m1*sa0a1;
  l1cm_m1_ca0a1 = s->l1cm_m1*ca0a1;
  l0_l1cm_m1_sa1 = s->l0_l1cm_m1*sa1;
  l0_l1cm_m1_ca1 = s->l0_l1cm_m1*ca1;
  l0cm_m0_p_l0_m1_sa0 = s->l0cm_m0_p_l0_m1*sa0;
  l0cm_m0_p_l0_m1_ca0 = s->l0cm_m0_p_l0_m1*ca0;

  expr1 = -s->g_l1cm_m1*sa0a1;

  d = s->expr2;
  c = d + l0_l1cm_m1_ca1;
  f = action[1] - s->viscous_friction[1]*a1d
    - s->force_forearm_fx*s->force_forearm_r*ca0a1
    - s->force_forearm_fz*s->force_forearm_r*sa0a1
    - ( expr1 + a0d_2*l0_l1cm_m1_sa1 );

  b = c;
  a = s->expr3 + b + l0_l1cm_m1_ca1;
  e = action[0] - s->viscous_friction[0]*a0d
    - s->force_upper_arm_fx*s->force_upper_arm_r*ca0
    - s->force_upper_arm_fz*s->force_upper_arm_r*sa0
    - s->force_forearm_fx*s->force_forearm_r*ca0a1
    - s->force_forearm_fz*s->force_forearm_r*sa0a1
    - s->force_forearm_fx*s->upper_arm_length*ca0
    - s->force_forearm_fz*s->upper_arm_length*sa0
    - ( expr1 - s->g_l0cm_m0_p_l0_m1*sa0
	- 2*a0d_a1d*l0_l1cm_m1_sa1 - a1d_2*l0_l1cm_m1_sa1 );
  
  determinant = a*d - b*c;
  a0dd = accelerations[0] = (d*e - b*f)/determinant;
  a1dd = accelerations[1] = (-c*e + a*f)/determinant;
}

/************************************************************************/
/* This is what is called on each integration step */

void integrate_one_time_step( SIM *s )
{ 
  int i;
  double new_velocities[N_DOFS];
  double accelerations[N_DOFS];

  if( s->status == CRASHED )
    {
      s->time += s->time_step;
      return;
    }

  forward_dynamics( s, s->torque, accelerations );

  // Trapezoidal integration
  for( i = 0; i < N_DOFS; i++ )
    {
      new_velocities[i] = s->state[i + N_DOFS]
	+ accelerations[i]*s->time_step;
      s->state[i] += 0.5*(new_velocities[i] 
			  + s->state[i + N_DOFS])*s->time_step;
      s->state[i + N_DOFS] = new_velocities[i];
    }
  s->time += s->time_step;

  for( i = 0; i < N_X; i++ )
    {
      if ( isnan( s->state[i] ) || isinf( s->state[i]) )
	{
	  fprintf( stderr, "Inf/Nan detected.\n" );
	  s->status = CRASHED;
	  return;
	}
    }

  forward_kinematics( s );
}

/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
